
/*===============
  Project History
  ===============

  Personnel:
  KB - Kai Bolay owner of the Zombiekeys extension
  AG - Lead Developer and owner of the Zombiekeys Mozdev project



  Version 1.0 - Released: 03/04/2006

    KB First Official Release, compatible with Firefox 1.5.* and Thunderbird 1.5.*


  Version 1.0.1 - Released: 06/11/2006

    KB Marked as compatible with Firefox 2.0


  Version 1.1.3 - Released: 03/05/2007

    KB support Thunderbird 2.0
    KB support SeaMonkey 1.1 / SuiteRunner 1.5a - thanks to Philip Chee
    KB added XBL binding to operate on html::input and derived elements - thanks to Arno Renevier
    KB French localization - thanks to Arno Renevie


  Version 1.1.4 - Released: 08/05/2007

    KB Support Thunderbird 2.0
    KB Support SeaMonkey 1.1 / SuiteRunner 1.5a - thanks to Philip Chee
    KB added XBL binding to operate on chrome://global/content/platformHTMLBindings.xml#inputFields and derived elements - thanks to Arno Renevier and Philip Chee
    KB French localization - thanks to Arno Renevier


  Version 1.3 - Released: 20/10/2010

    bumped compatibility to TB 3.1.*, FX 3.6 (27/06/2010)
    AG Added Postbox 1.1 support
    AG New Addon Icon.


  Version 2.0 - Released: 20/01/2011

    AG Adds compatibility to Firefox 4 and Postbox 2.1.
    AG I also redesigned the icon, and a 64 pixel version for the new detail view in Firefox 4
    AG Added lightning support for new event window


  Version 2.1 - Released: 05/07/2011

    AG Compatibility Update for Fx version 5


  Version 2.5 - Work in Progress

    AG Bumped up max Compatibility for Thunderbird and Firefox to 6.* to better align with Mozilla's new rapid release cycle
    AG Added Zoimbiekeys Toolbar button
    AG Added multiple keyboard layout support
    AG Added Configuration dialog
    Known issue: French localization is incomplete
    Known issue: The keyboard shortcuts on the menus do not reflect the shortcuts of the localized layouts 100% accurately (to add the shortcut descriptions dynamically is planned for a future release)


  */

var ZombieKeys = new function() {
	var debugLevel = 0, mAppver = null, mAppName = null;

	// modifiers: [ctrl, shift, alt]
	// make deadKeys accessible from outside (this. instead of var)
	// additional unicode keys from http://www.fileformat.info/info/unicode/
	// ideas for additional diacritics support from http://dry.sailingissues.com/us-international-keyboard-layout.html
	var deadKeys= [
				{	"modifiers": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id" : 1,
					"keyCode": 192, // Grave "`" US=192
					"mapping": {"a": "\u00e0",
								"e": "\u00e8",
								"i": "\u00ec",
								"o": "\u00f2",
								"u": "\u00f9",
								"A": "\u00c0",
								"E": "\u00c8",
								"I": "\u00cc",
								"O": "\u00d2",
								"U": "\u00d9"},
					"menuFake": false},

				{	"modifiers": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id": 2,
					"keyCode": 222, // Acute "'"
					"mapping": {"a": "\u00e1",
								"e": "\u00e9",
								"i": "\u00ed",
								"o": "\u00f3",
								"u": "\u00fa",
								"y": "\u00fd",
								"A": "\u00c1",
								"E": "\u00c9",
								"I": "\u00cd",
								"O": "\u00d3",
								"U": "\u00da",
								"Y": "\u00dd",
								"d": "\u00f0",
								"D": "\u00d0"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 3,
					"keyCode": 54, // Circumflex "^" us:54   uk:54
					"mapping": {"A": "\u00c2", "a": "\u00e2",
								"E": "\u00ca", "e": "\u00ea",
								"I": "\u00ce", "i": "\u00ee",
								"O": "\u00d4", "o": "\u00f4",
								"U": "\u00db", "u": "\u00fb"
								},
					"menuFake": false},

					// can not user CTRL+SHIFT+# on UK keyboard as it switches tab groups in Fx 5.0
				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 4,
					"keyCode": 192, // Tilde "~"  uk:222  us:192
					"mapping": {"A": "\u00c3", "a": "\u00e3",
								"N": "\u00d1", "n": "\u00f1",
								"O": "\u00d5", "o": "\u00f5"
								},
					"menuFake": false},

				{	"modifiers": {"ctrlKey": true, "shiftKey":	true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 5,
					"keyCode": 59, // Umlaut ":"   uk:59  us:59
					"mapping": {"a": "\u00e4",
								"e": "\u00eb",
								"i": "\u00ef",
								"o": "\u00f6",
								"u": "\u00fc",
								"y": "\u00ff",
								"A": "\u00c4",
								"E": "\u00cb",
								"I": "\u00cf",
								"O": "\u00d6",
								"U": "\u00dc",
								"Y": "\u0178"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 6,
					"keyCode": 50, // Ring "@"
					"mapping": {"a": "\u00e5",
								"A": "\u00c5",
								"u": "\u016f",
								"U": "\u016e",
								"y": "\u1e99",
								"w": "\u1e98"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 7,
					"keyCode": 55, // AE,OE,Sharp S "&"   us=55,  uk=55
					"mapping": {"a": "\u00e6",
								"A": "\u00c6",
								"o": "\u0153",
								"O": "\u0152",
								"s": "\u00df"},
					"menuFake": false},


				{	"modifiers":   {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id": 8,
					"keyCode": 191, // Stroke "/"
					"mapping": {"o": "\u00f8",
								"O": "\u00d8",
								"c": "\u00a2"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id": 9,
					"keyCode": 190, // Caron (Hacek) ">"
					"mapping": {"A": "\u01cd", "a": "\u01ce",
								"C": "\u010c", "c": "\u010d",
								"D": "\u010e", "d": "\u010f",
								"E": "\u011a", "e": "\u011B",
								"G": "\u01e6", "g": "\u01e7",
								"H": "\u021e", "h": "\u021f",
								"I": "\u01cf", "i": "\u01d0",
								"j": "\u01f0",
								"K": "\u01e8", "k": "\u01e9",
								"L": "\u013d", "l": "\u013e",
								"N": "\u0147", "n": "\u0148",
								"O": "\u01d1", "o": "\u01d2",
								"R": "\u0158", "r": "\u0159",
								"S": "\u0160", "s": "\u0161",
								"T": "\u0164", "t": "\u0165",
								"U": "\u01d3", "u": "\u01d4",
								"Z": "\u017d", "z": "\u017e",
								" ": "\u02c7"},
					"menuFake": false},

					{"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 10,
					"keyCode": 190, // Cedilla ">"
					"mapping": {
								"C": "\u00c7", "c": "\u00e7",
								"D": "\u1e10", "d": "\u1e11",
								"G": "\u0122", "g": "\u0123",
								"K": "\u0136", "k": "\u0137",
								"N": "\u0145", "n": "\u0146",
								"R": "\u0156", "r": "\u0157",
								"S": "\u015e", "s": "\u015f",
								"T": "\u0162", "t": "\u0163",
								" ": "\u00B8"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 11,
					"keyCode": 57, // Breve "("
					"mapping": {
								"A": "\u0102", "a": "\u0103",
								"E": "\u0114", "e": "\u0115",
								"G": "\u011e", "g": "\u011f",
								"H": "\u1e2a", "h": "\u1e2b", // breve below
								"I": "\u012c", "i": "\u012d",
								"O": "\u014e", "o": "\u014f",
								"U": "\u016c", "u": "\u016d",
								"X": "\u04c1", "x": "\u04c2",  // ZHE
								" ": "\u02D8"},
					"menuFake": false},

				{	"modifiers": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 12,
					"keyCode": 56, // Ogonek "*"
					"mapping": {
								"A": "\u0104", "a": "\u0105",
								"E": "\u0118", "e": "\u0119",
								"I": "\u012e", "i": "\u012f",
								"O": "\u01ea", "o": "\u01eb",
								"U": "\u0172", "u": "\u0173",
								" ": "\u02Db"},
					"menuFake": false},

				{	"modifiers": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 13,
					"keyCode": 109, // Macron "_"
					"mapping": {
								"A": "\u0100", "a": "\u0101",
								"E": "\u0112", "e": "\u0113",
								"I": "\u012a", "i": "\u012b",
								"O": "\u014c", "o": "\u014d",
								"U": "\u016a", "u": "\u016b",
								"Y": "\u0232", "y": "\u0233",
								" ": "\u00af"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id": 14,
					"keyCode": 190, // Overdot "."
					"mapping": {
								"A": "\u0226", "a": "\u0227",
								"C": "\u010a", "c": "\u010b",
								"E": "\u0116", "e": "\u0117",
								"I": "\u0130",
								"M": "\u1e40", "m": "\u1e41",
								"N": "\u1e44", "n": "\u1e45",
								"O": "\u022e", "o": "\u022f",
								"P": "\u1e56", "p": "\u1e57",
								"R": "\u1e58", "r": "\u1e59",
								"S": "\u1e60", "s": "\u1e61",
								"W": "\u1e86", "w": "\u1e87",
								"X": "\u1e8a", "x": "\u1e8b",
								"Y": "\u1e8e", "y": "\u1e8f",
								"Z": "\u017b", "z": "\u017c"
								},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey": false},
					"id": 15,
					"keyCode": 59, // Underdot ";"
					"mapping": {
								"A": "\u1ea0", "a": "\u1ea1",
								"D": "\u1e0c", "d": "\u1e0d",
								"E": "\u1eb8", "e": "\u1eb9",
								"H": "\u1e24", "h": "\u1e25",
								"I": "\u1eca", "i": "\u1ecb",
								"K": "\u1e32", "k": "\u1e33",
								"L": "\u1e36", "l": "\u1e37",
								"O": "\u1ecc", "o": "\u1ecd",
								"S": "\u1e62", "s": "\u1e63",
								"T": "\u1e6c", "t": "\u1e6d",
								"U": "\u1ee4", "u": "\u1ee5",
								"V": "\u1e7e", "v": "\u1e7f",
								"W": "\u1e88", "w": "\u1e89",
								"Y": "\u1ef4", "y": "\u1ef5",
								" ": "\u0323"},
					"menuFake": false},

				{	"modifiers":   {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"defaultmods": {"ctrlKey": true, "shiftKey": true, "altKey": false},
					"id": 16,
					"keyCode": 50, // Double Accute """
					"mapping": {
								"O": "\u0150","o": "\u0151",
								"U": "\u0170","u": "\u0171",
								" ": "\u02dd"
								},
					"menuFake": false}

					];


	var aliveKeys = [
					{ "id": 1,
					  "modifiers":   {"ctrlKey": true, "shiftKey":  true, "altKey":  true},
					  "defaultmods": {"ctrlKey": true, "shiftKey":  true, "altKey":  true},
					  "charCode": 0,
					  "keyCode": 191, // "?"
					  "target": "\u00bf"},

					{ "id": 2,
					  "modifiers":   {"ctrlKey": true, "shiftKey":  true, "altKey":  true},
					  "defaultmods": {"ctrlKey": true, "shiftKey":  true, "altKey":  true},
					  "charCode": 33,
					  "keyCode": 49, // "!"
					  "target": "\u00a1"},

					{ "id": 3,
					  "modifiers":   {"ctrlKey": true, "shiftKey": false, "altKey":  true},
					  "defaultmods": {"ctrlKey": true, "shiftKey": false, "altKey":  true},
					  "charCode": 36,
					  "keyCode": 52, // "@" us=charCode101 "e" - not necessary on Irish keyboard - it has Euro symbol on ALT-GR + 4
					                 // on UK lets put it on the $ key (36)
					  "target": "\u20ac"}];

	// Locale Specific Layouts
	// we can add optional shiftKey overrides to cater for the locale!
	var layouts = [{ "locale" : "us_int",
					 "map_deadKeys" : [
						{"id" : 1, "keyCode": 192}, // "`"
						{"id" : 2, "keyCode": 222}, // "'"
						{"id" : 3, "keyCode":  54}, // "^"
						{"id" : 4, "keyCode": 192}, // "~"
						{"id" : 5, "keyCode":  59}, // ":"
						{"id" : 6, "charCode":  0, "keyCode": 192}, // "`"
						{"id" : 7, "keyCode":  55}, // "&"
						{"id" : 8, "keyCode": 188}, // ","
						{"id" : 9, "keyCode": 191}  // "/"
					 ],
					 "map_liveKeys" : [
						 {"id": 1, "charCode":   0, "keyCode": 191}, // "?"
						 {"id": 2, "charCode":  33, "keyCode":  49}, // "!"
						 {"id": 3, "charCode": 101, "keyCode":  52}  // "@"
					  ]
					},

					{ "locale" : "us_dvorak",
    				  "map_deadKeys" : [
						{"id" : 1, "keyCode": 192}, // "`"
						{"id" : 2, "keyCode": 222}, // "'"
						{"id" : 3, "keyCode":  54}, // "^"
						{"id" : 4, "keyCode": 192}, // "~"
						{"id" : 5, "keyCode":  59}, // ":"
						{"id" : 6, "charCode":0, "keyCode": 192}, // "`"
						{"id" : 7, "keyCode":  55}, // "&"
						{"id" : 8, "keyCode": 188}, // ","
						{"id" : 9, "keyCode": 191} // "/"
					  ],
					  "map_liveKeys" : [
						 {"id": 1, "charCode":   0, "keyCode": 191}, // "?"
						 {"id": 2, "charCode":  33, "keyCode":  49}, // "!"
						 {"id": 3, "charCode": 101, "keyCode":  52}  // "@"
					  ]
					},

					{ "locale" : "irl",
					  "map_deadKeys" : [
						{"id" : 1, "keyCode": 192}, // "'"
						{"id" : 2, "keyCode": 222}, // "#"
						{"id" : 3, "keyCode":  54}, // "^"
						{"id" : 4, "keyCode": 222, "altkey": true}, // "~" // avoid tab switch in Fx5
						{"id" : 5, "keyCode":  59}, // ":"
						{"id" : 6, "keyCode": 192}, // "@" "charCode": 64,
						{"id" : 7, "keyCode":  55}, // "&"
						{"id" : 8, "keyCode": 188}, // ","
						{"id" : 9, "keyCode": 191}  // "/"
					  ],
					  "map_liveKeys" : [
						 {"id": 1, "charCode":   0, "keyCode": 191}, // "?"
						 {"id": 2, "charCode":  33, "keyCode":  49}, // "!"
						 {"id": 3, "charCode": 101, "keyCode":  54}  // "4"
					  ]
					},

					{ "locale" : "uk",
    				  "map_deadKeys" : [
						{"id" : 1, "keyCode": 192}, // "`"
						{"id" : 2, "keyCode": 222}, // "'"
						{"id" : 3, "keyCode":  54}, // "^"
						{"id" : 4, "keyCode": 222, "altkey": true}, // "~"
						{"id" : 5, "keyCode":  59}, // ":"
						{"id" : 6, "keyCode": 192}, // "@"
						{"id" : 7, "keyCode":  55}, // "&"
						{"id" : 8, "keyCode": 188}, // ","
						{"id" : 9, "keyCode": 191}  // "/"
					  ],
					  "map_liveKeys" : [
						 {"id": 1, "charCode":   0, "keyCode": 191}, // "?"
						 {"id": 2, "charCode":  33, "keyCode":  49}, // "!"
						 {"id": 3, "charCode": 101, "keyCode":  36}  // "$"
					  ]
					}

				 ];


	var hexUnicodeKey = {"modifiers": {"ctrlKey": false, "shiftKey": false,
									   "altKey":  true},
						 "keyCode": 88}; // "x"

	var decUnicodeModifiers = {"ctrlKey": false, "shiftKey": false,
							   "altKey":  true};

	var decUnicode = "";

	var keyEventProps = ['keyCode', 'charCode', 'shiftKey',
						 'altKey', 'ctrlKey'];

	function log(message) {
		var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
			.getService(Components.interfaces.nsIConsoleService);
		consoleService.logStringMessage("[ZombieKeys] " + message + (message != "\n" ? " " : ""));
	};

	function logKey(why, event) {
		var message=why;
		for (var k=0; k<keyEventProps.length; k++) {
			message += keyEventProps[k] + "=" + event[keyEventProps[k]]+" ";
		}
		log(message+"\n");
	}

	function logMappingString(mapping) {
		var message="";
		for (var mappee in mapping) {
			message += mappee+" ("+mappee.charCodeAt(0)+") => "+
				mapping[mappee]+" ("+mapping[mappee].charCodeAt(0)+") ";
		}
		return message+"\n";
	}

	function checkModifiers(event, modifiers) {
		for (var key in modifiers) {
			if (event[key] != modifiers[key]) {
				return false;
			}
		}
		return true;
	}

	function fakeKey(event, chr, createKeyUp) {
		try {
			if (debugLevel > 0) {logKey("fakeKey: ('" +chr + "' " + event.type + ")", event);}
			var view = window; // event.view;
			var target = event.target ? event.target : document.commandDispatcher.focusedElement;

			if(!target && debugLevel>0) debugger;

			if (target.nodeName=="HTML") {


	// 			if (debugLevel > 0) {logKey("faking '"+chr+"': " + keydown_event.type + " + target: " + target.id + "\n", keydown_event);}
	// 			target.dispatchEvent(keydown_event);
	//
				// PRESS
				var keypress_event = document.createEvent("KeyboardEvent"); // KeyEvents
				keypress_event.initKeyEvent("keypress", true, true, null,
							   false, false, false, false,
							   0, chr.charCodeAt(0));
				if (debugLevel > 0) {logKey("faking '"+chr+"': " + keypress_event.type + " + target: " + target.id + "\n", keypress_event);}
				target.dispatchEvent(keypress_event);
			}
			else {
				insertAtCaret(target, chr);
			}
			// "keypress" or "keyup"
			event.preventDefault();
			event.stopPropagation();


		}
		catch(ex) {
			log("exception during fakekey: " + ex.toString());
		}

	};

	// get string preceeding cursor
	function getPreceedingString(elt, numchars) {
		var pos = elt.selectionStart;
		if (pos < numchars) {numchars=pos;}
		return elt.value.slice(pos-numchars,pos);
	}

	// replace string preceeding cursor
	function replacePreceedingString(elt, numchars, newtext) {
		var posStart = elt.selectionStart;
		var posEnd = elt.selectionEnd;
		var scrollTop = elt.scrollTop;
		var scrollLeft = elt.scrollLeft;

		if (posStart < numchars) {numchars=posStart;}

		elt.value = elt.value.slice(0,posStart-numchars)+
					newtext+
					elt.value.slice(posEnd);
		var newpos = posStart+newtext.length;
		elt.selectionStart = newpos;
		elt.selectionEnd = newpos;

		elt.scrollTop = scrollTop;
		elt.scrollLeft = scrollLeft;
		elt.focus();
	}


	// modify the key mappings acording to the selected locale
	this.initLocale = function(locale) {
		log("Initialize key locale: " + locale + " ...");
		var theLayout; // determine locale
		Array.forEach(layouts, function(layout) {
			if (layout.locale==locale) {
				theLayout = layout;
			}
		});


		for (var k=0; k<deadKeys.length; k++) {
			Array.forEach(theLayout.map_deadKeys, function(Map) {
				if(Map.id == deadKeys[k].id) {
					deadKeys[k].keyCode = theLayout.map_deadKeys[k].keyCode;
					// Special mods
					if (typeof theLayout.map_deadKeys[k].charCode)
						deadKeys[k].charCode = theLayout.map_deadKeys[k].charCode;
					else
						if (layouts[0].map_deadKeys[k].charCode)
							deadKeys[k].charCode = layouts[0].map_deadKeys[k].charCode; // us_int as default

					if (typeof theLayout.map_deadKeys[k].shiftKey !== "undefined")
						deadKeys[k].modifiers.shiftKey = theLayout.map_deadKeys[k].shiftKey;
					else
						deadKeys[k].modifiers.shiftKey = deadKeys[k].defaultmods.shiftKey;

					if (typeof theLayout.map_deadKeys[k].altKey !== "undefined")
						deadKeys[k].modifiers.altKey = theLayout.map_deadKeys[k].altKey;
					else
						deadKeys[k].modifiers.altKey = deadKeys[k].defaultmods.altKey;
				}
			});
		}

		for (var k=0; k<aliveKeys.length; k++) {
			Array.forEach(theLayout.map_liveKeys, function(Map) {
				if(Map.id == aliveKeys[k].id) {
					aliveKeys[k].keyCode = Map.keyCode;
					aliveKeys[k].charCode = Map.charCode;

					if (typeof Map.ctrlKey !== "undefined" )
						aliveKeys[k].modifiers.ctrlKey = Map.ctrlKey;
					else
						aliveKeys[k].modifiers.ctrlKey = aliveKeys[k].defaultmods.ctrlKey;

					if (typeof Map.shiftKey !== "undefined" )
						aliveKeys[k].modifiers.shiftKey = Map.shiftKey;
					else
						aliveKeys[k].modifiers.shiftKey = aliveKeys[k].defaultmods.shiftKey;

					if (typeof Map.altKey !== "undefined" )
						aliveKeys[k].modifiers.altKey = Map.altKey;
					else
						aliveKeys[k].modifiers.altKey = aliveKeys[k].defaultmods.altKey;
				}
			});
		}
	}


	function keyPressHandler(event) {
		if (debugLevel > 5) {logKey("press: ", event);}

		// check "alive" keys which fire during "keypress"
		for (var k=0; k<aliveKeys.length; k++) {
			if (  ( (event.keyCode && event.keyCode  == aliveKeys[k].keyCode)
					||
				    (event.charCode && event.charCode == aliveKeys[k].charCode)    )
				&&
				checkModifiers(event, aliveKeys[k].modifiers)) {
				fakeKey(event, aliveKeys[k].target);
				return;
			}
		}

		// check if/which "dead" key is active
		for (var k=0; k<deadKeys.length; k++) {
			if (deadKeys[k].alive) {
				deadKeys[k].alive = false;
				// check if there's a mapping for this character
				// entered after "dead" key
				if (debugLevel > 10) {
					log("detected zombie # "+k+" for "+event.charCode+
						" => " + logMappingString(deadKeys[k].mapping));
				}
				var code = String.fromCharCode(event.charCode);
				if (deadKeys[k].mapping[code]) {
					if (deadKeys[k].menuFake && event.type=='keypress') {
						// make it a pair of events
						deadKeys[k].menuFake = false;
						fakeKey(event, deadKeys[k].mapping[code], true); // copy event with keyup!
					}
					else
						fakeKey(event, deadKeys[k].mapping[code], false);
					return;
				}
			}
		}
	};

	function keyUpHandler(event) {
		if (debugLevel > 5) {logKey("up: ", event);}

		// check "alive" keys which only fire on "keyup"
		for (var k=0; k<aliveKeys.length; k++) {
			if (aliveKeys[k].keyCode &&
				(event.keyCode == aliveKeys[k].keyCode) &&
				checkModifiers(event, aliveKeys[k].modifiers)) {
				fakeKey(event, aliveKeys[k].target);
				return;
			}
		}

		// check if "dead" key is activated
		for (var k=0; k<deadKeys.length; k++) {
			if (
				(
				 (event.keyCode == deadKeys[k].keyCode)
				 ||
				 (deadKeys[k].charCode && event.charCode == deadKeys[k].charCode)
				)
				&&
				checkModifiers(event, deadKeys[k].modifiers)) {
				if (debugLevel > 10) {
					log("created zombie # "+k+": "+
					    logMappingString(deadKeys[k].mapping));
					debugger;
				}
				deadKeys[k].alive = true;
				event.preventDefault();
				event.stopPropagation();
				return;
			}
		}

		// process ALT+X (treat preceeding 4 chars as hex unicode string
		// and replace with corresponding char)
		if ((event.keyCode == hexUnicodeKey.keyCode) &&
			checkModifiers(event, hexUnicodeKey.modifiers)) {
			var hexUnicode = getPreceedingString(event.target, 4);
			var code = parseInt(hexUnicode, 16);
			replacePreceedingString(event.target, 4, String.fromCharCode(code));
			return;
		}

		// process ALT+<numeric keypad>
		if (checkModifiers(event, decUnicodeModifiers)) {
			if ((event.keyCode >= 96) && (event.keyCode <= 105)) {
				decUnicode += (event.keyCode-96).toString(); // add digit
				event.preventDefault();
				event.stopPropagation();
				return;
			} else if (event.keyCode == 18) {
				// released ALT - let's insert
				if (decUnicode != "") {
					var code = parseInt(decUnicode, 10);
					fakeKey(event, String.fromCharCode(code));
					decUnicode = "";
				}
				event.preventDefault();
				event.stopPropagation();
				return;
			}
		}
	};


	function insertAtCaret(element, chr) {
		if (!element) {
			log("insertAtCaret for chr '" + chr.toString() + "' cannot be done - no element passed!");
			return;
		}
		if (element.nodeName=="HTML") { // Composer Windows
		    if (window.getSelection && window.getSelection().getRangeAt) {
		        range = window.getSelection().getRangeAt(0);
		        node = range.createContextualFragment(chr.toString());
		        range.insertNode(node);
		    } else if (document.selection && document.selection.createRange) {
		        document.selection.createRange().pasteHTML(chr.toString());
		    }
		}
		else {
			var start = element.selectionStart, end = element.selectionEnd;
			element.value = element.value.substring(0, start) + chr + element.value.substring(end, element.value.length);
			element.selectionStart = start+1;
			element.selectionEnd = start+1;
		}
	}



	this.Options = {
		selectItem: function(layout) {
			var lst = document.getElementById('Layouts');
			if (lst) {
				for (var i=0; i<lst.itemCount; i++) {
					var item = lst.getItemAtIndex(i);
					if (item && item.value == layout) {
						lst.selectItem (item);
						break;
					}
				}
			}
		},

		selectLayout: function(layout, win) {
			var img = document.getElementById('layoutImage');
			if (img)
				img.className = "layout-" + layout;


		}
	}

	this.viewOptions = function () {
		var oldLocale = this.getCurrentLocale();
		window.openDialog('chrome://zombiekeys/content/zombiekeys-options.xul',
			'zombiekeys-options','chrome,titlebar,centerscreen,modal,resizable,innerWidth=440,innerHeight=320',ZombieKeys).focus();
		var newLocale = this.getCurrentLocale();
		// for safety, force to initialize locale after options are closed
		log("old locale=" + oldLocale + " - initialize new locale=" + newLocale + "...");
		this.initLocale(newLocale);
	}

	this.getStringBundle = function() {
		var zk_bundle =
			Components.classes["@mozilla.org/intl/stringbundle;1"]
				.getService(Components.interfaces.nsIStringBundleService)
				.createBundle("chrome://zombiekeys/locale/zombiekeys.properties");
		return zk_bundle;
	}

	this.getResourceString = function(id) {
		try {
			var txt = ZombieKeys.getStringBundle().GetStringFromName('extensions.zombiekeys.' + id);
			return txt;
		}
		catch (ex) {
			return 'Missing Resource string {extensions.zombiekeys.' + id + '}\n'
					+ ex.toString();
		}
	}

	this.showHowToUse = function() {

		var text = ZombieKeys.getResourceString('howToUse');
		alert (text);
	}

	this.loadOptions = function () {
		ZombieKeys.Options.selectItem(this.getCurrentLocale());
	}

	this.acceptOptions = function () {
		var lst = document.getElementById('Layouts');
		if (lst) {
			var locale = lst.selectedItem.value;
			Components.classes["@mozilla.org/preferences-service;1"]
				.getService(Components.interfaces.nsIPrefBranch)
				.setCharPref('extensions.zombiekeys.currentLayout', locale);
		}

	}

	this.closeOptions = function () {
		log("Current locale remains=" + this.getCurrentLocale());
	}

	this.getCurrentLocale = function () {
		return Components.classes["@mozilla.org/preferences-service;1"]
 					.getService(Components.interfaces.nsIPrefBranch)
 					.getCharPref('extensions.zombiekeys.currentLayout');
	}

	this.setDebugLevel = function(level) {
		debugLevel = level;
		Components.classes["@mozilla.org/preferences-service;1"]
				.getService(Components.interfaces.nsIPrefBranch)
				.setIntPref('extensions.zombiekeys.debugLevel', level);
	}



	this.init = function(that) {
		that.debugLevel = Components.classes["@mozilla.org/preferences-service;1"]
				.getService(Components.interfaces.nsIPrefBranch)
				.getIntPref('extensions.zombiekeys.debugLevel');
		//if (debugLevel > 0) {
			log("Starting up with debug level=" + this.debugLevel + "\nSet to 0 to avoid lots of console messages from ZombieKeys");
		//}
		var locale = that.getCurrentLocale();
		if (that.debugLevel > 0)
			log("Call initLocale (" + locale + ") ...");

		that.initLocale(locale);
		if (window) {
			window.addEventListener('keypress', keyPressHandler, false);
			window.addEventListener('keyup',	keyUpHandler,	 false);
		}


	}

	this.showPopup = function(button, popupId) {
		var p =  document.getElementById(popupId);
		if (p) {
				document.popupNode = button;
				if (typeof p.openPopup=='undefined')
						p.showPopup(button, -1,-1,"context","bottomleft","topleft");
				else
						p.openPopup(button,'after_start', -1,-1,"context",false);
		}
	}

	this.awakeZombie = function (id) {
		for (var k=0; k<deadKeys.length; k++) {
			if (deadKeys[k].id === id)
			{
				if (debugLevel > 0) {
					log("awakeZombie(" + id + ") matched keyCode: " + deadKeys[k].keyCode + "\n\nZombiefy the dead key...");
					debugger;
				}
				deadKeys[k].menuFake = true;
				deadKeys[k].alive = true; // activate modifiers
				break;
			}
		}
	}

	this.fakeAliveKey = function(id)
	{
		var target = document.commandDispatcher.focusedElement;
		// initKeyEvent(type, canBubble, cancelable, view,
		//				ctrlKey, altKey, shiftKey, metaKey, keyCode, charCode);
		// DOWN
		for (var k=0; k<aliveKeys.length; k++) {
			if (aliveKeys[k].id ===id)
				var theKey=aliveKeys[k];
		}


		var keydown_event = document.createEvent("KeyboardEvent"); // KeyEvents
		keydown_event.initKeyEvent("keydown", true, true, null,
							 	theKey.modifiers.ctrlKey,
							 	theKey.modifiers.altKey,
							 	theKey.modifiers.shiftKey,
							 	false,
							 	theKey.keyCode,
							 	theKey.charCode);


		if(theKey)
		{
			insertAtCaret(target, theKey.target);
			return;
		}

		keydown_event.preventDefault();
		keydown_event.stopPropagation();

// UP
// 		var keyup_event = document.createEvent("KeyboardEvent");
// 		keyup_event.initKeyEvent("keyup", true, true, null,
// 							 	ctrl, alt, shift, false,
// 							 	keyCode, charCode); // .charCodeAt(0)
// 		target.dispatchEvent(keyup_event);
// 		if (debugLevel > 0) {logKey("faking '"+chr+"': " + keyup_event.type + " + target: " + target.id + "\n", keyup_event);}
//


    }


	this.showContentTab = function(URL) {
		try {
			var sTabMode="";
			var tabmail;
			tabmail = document.getElementById("tabmail");
			if (!tabmail) {
				// Try opening new tabs in an existing 3pane window
				var mail3PaneWindow = Components.classes["@mozilla.org/appshell/window-mediator;1"]
										 .getService(Components.interfaces.nsIWindowMediator)
										 .getMostRecentWindow("mail:3pane");
				if (mail3PaneWindow) {
					tabmail = mail3PaneWindow.document.getElementById("tabmail");
					mail3PaneWindow.focus();
				}
			}
			if (tabmail) {
				sTabMode = (this.Application() == "Thunderbird" && this.Appver()>=3) ? "contentTab" : "3pane";
				tabmail.openTab(sTabMode,
				{contentPage: URL, clickHandler: "specialTabs.siteClickHandler(event, '');"});
			}
			else
				window.openDialog("chrome://messenger/content/", "_blank",
								  "chrome,dialog=no,all", null,
			  { tabType: "contentTab", tabParams: {contentPage: URL, clickHandler: "specialTabs.siteClickHandler(event, '');", id:"Zombiekeys_Weblink"} } );
		}
		catch(e) { return false; }
		return true;

	}


	this.AppverFull = function() {
		var appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
						.getService(Components.interfaces.nsIXULAppInfo);
		return appInfo.version;
	}

	this.Appver = function() {
		if (null == mAppver) {
		var appVer=this.AppverFull().substr(0,3); // only use 1st three letters - that's all we need for compatibility checking!
			mAppver = parseFloat(appVer); // quick n dirty!
		}
		return mAppver;
	}

	this.Application = function() {
		if (null==mAppName) {
		var appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
						.getService(Components.interfaces.nsIXULAppInfo);
			const FIREFOX_ID = "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}";
			const THUNDERBIRD_ID = "{3550f703-e582-4d05-9a08-453d09bdfdc6}";
			const SEAMONKEY_ID = "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}";
			const POSTBOX_ID = "postbox@postbox-inc.com";
			switch(appInfo.ID) {
				case FIREFOX_ID:
					return mAppName='Firefox';
				case THUNDERBIRD_ID:
					return mAppName='Thunderbird';
				case SEAMONKEY_ID:
					return mAppName='SeaMonkey';
				case POSTBOX_ID:
					return mAppName='Postbox';
				default:
					mAppName=appInfo.name;
					return appInfo.name;
			}
		}
		return mAppName;
	}

};



if (window) {
	window.addEventListener("load", function() {setTimeout(
	  function () {ZombieKeys.init(ZombieKeys);}
	,3000
	);}, false);
}
